@@ -40,6 +40,13 @@ struct module
 	void *ieee1394_table;
 	unsigned int ccw_size;
 	void *ccw_table;
+	unsigned int pnp_size;
+	void *pnp_table;
+	unsigned int pnp_card_size;
+	unsigned int pnp_card_offset;
+	void *pnp_card_table;
+	unsigned int input_size;
+	void *input_table;
 
 	/* File contents and length. */
 	void *data;
