@@ -119,14 +119,8 @@ int main(int argc, char *argv[])
 	/* Rest is options */
 	for (i = 2; i < argc; i++) {
 		options = realloc(options,
-				  strlen(options) + 2 + strlen(argv[i]) + 2);
-		/* Spaces handled by "" pairs, but no way of escaping
-                   quotes */
-		if (strchr(argv[i], ' '))
-			strcat(options, "\"");
+				  strlen(options) + 1 + strlen(argv[i]) + 1);
 		strcat(options, argv[i]);
-		if (strchr(argv[i], ' '))
-			strcat(options, "\"");
 		strcat(options, " ");
 	}
 
