@@ -39,6 +39,9 @@
 #include <sys/wait.h>
 #include <syslog.h>
 
+#define streq(a,b) (strcmp((a),(b)) == 0)
+#define ARRAY_SIZE(x) (sizeof(x) / sizeof((x)[0]))
+
 #include "zlibsupport.h"
 #include "list.h"
 #include "backwards_compat.c"
@@ -60,21 +63,16 @@ typedef void (*errfn_t)(const char *fmt, ...);
 
 /* Do we use syslog or stderr for messages? */
 static int log;
-/* Should we be totally quiet? */
-static int quiet = 0;
 
 static void message(const char *prefix, const char *fmt, va_list *arglist)
 {
 	char *buf, *buf2;
 
-	if (quiet)
-		return;
-
 	vasprintf(&buf, fmt, *arglist);
 	asprintf(&buf2, "%s%s", prefix, buf);
 
 	if (log)
-		syslog(LOG_NOTICE, buf2);
+		syslog(LOG_NOTICE, "%s", buf2);
 	else
 		fprintf(stderr, "%s", buf2);
 	free(buf2);
@@ -119,8 +117,10 @@ static void *do_nofail(void *ptr, const char *file, int line, const char *expr)
 static void print_usage(const char *progname)
 {
 	fprintf(stderr,
-		"Usage: %s [--verbose|--version|--config|--remove] filename options\n",
-		progname);
+		"Usage: %s [-v] [-V] [-C config-file] [-n] [-i] [-q] [-o <modname>] <modname> [parameters...]\n"
+		"%s -r [-n] [-i] [-v] <modulename> ...\n"
+		"%s -l -t <dirname> [ -a <modulename> ...]\n",
+		progname, progname, progname);
 	exit(1);
 }
 
@@ -196,6 +196,29 @@ static void filename2modname(char *modname, const char *filename)
 	modname[i] = '\0';
 }
 
+static int lock_file(const char *filename)
+{
+	int fd = open(filename, O_RDWR, 0);
+
+	if (fd >= 0) {
+		struct flock lock;
+		lock.l_type = F_WRLCK;
+		lock.l_whence = SEEK_SET;
+		lock.l_start = 0;
+		lock.l_len = 1;
+		fcntl(fd, F_SETLKW, &lock);
+	} else
+		/* Read-only filesystem?  There goes locking... */
+		fd = open(filename, O_RDONLY, 0);
+	return fd;
+}
+
+static void unlock_file(int fd)
+{
+	/* Valgrind is picky... */
+	close(fd);
+}
+
 static void add_module(char *filename, int namelen, struct list_head *list)
 {
 	struct module *mod;
@@ -282,12 +305,12 @@ static void read_depends(const char *dirname,
 			 const char *start_name,
 			 struct list_head *list)
 {
-	char modules_dep_name[strlen(dirname) + sizeof("modules.dep") + 1];
+	char *modules_dep_name;
 	char *line;
 	FILE *modules_dep;
 	int done = 0;
 
-	sprintf(modules_dep_name, "%s/%s", dirname, "modules.dep");
+	asprintf(&modules_dep_name, "%s/%s", dirname, "modules.dep");
 	modules_dep = fopen(modules_dep_name, "r");
 	if (!modules_dep)
 		fatal("Could not load %s: %s\n",
@@ -300,6 +323,7 @@ static void read_depends(const char *dirname,
 		free(line);
 	}
 	fclose(modules_dep);
+	free(modules_dep_name);
 }
 
 /* We use error numbers in a loose translation... */
@@ -586,10 +610,14 @@ add_options(const char *modname,
 	    struct module_options *options)
 {
 	struct module_options *new;
+	char *tab; 
 
 	new = NOFAIL(malloc(sizeof(*new)));
 	new->modulename = NOFAIL(strdup(modname));
 	new->options = NOFAIL(strdup(option));
+	/* We can handle tabs, kernel can't. */
+	for (tab = strchr(new->options, '\t'); tab; tab = strchr(tab, '\t'))
+		*tab = ' ';
 	new->next = options;
 	return new;
 }
@@ -614,7 +642,7 @@ static const char *find_command(const char *modname,
 				const struct module_command *commands)
 {
 	while (commands) {
-		if (strcmp(commands->modulename, modname) == 0)
+		if (fnmatch(commands->modulename, modname, 0) == 0)
 			return commands->command;
 		commands = commands->next;
 	}
@@ -643,6 +671,19 @@ static char *add_extra_options(const char *modname,
 	return optstring;
 }
 
+/* If we don't flush, then child processes print before we do */
+static void verbose_printf(int verbose, const char *fmt, ...)
+{
+	va_list arglist;
+
+	if (verbose) {
+		va_start(arglist, fmt);
+		vprintf(fmt, arglist);
+		fflush(stdout);
+		va_end(arglist);
+	}
+}
+
 /* Do an install/remove command. */
 static void do_command(const char *modname,
 		       const char *command,
@@ -652,27 +693,21 @@ static void do_command(const char *modname,
 {
 	int ret;
 
-	if (verbose)
-		printf("%s %s\n", type, command);
+	verbose_printf(verbose, "%s %s\n", type, command);
 	if (dry_run)
 		return;
 
+	setenv("MODPROBE_MODULE", modname, 1);
 	ret = system(command);
 	if (ret == -1 || WEXITSTATUS(ret))
 		error("Error running %s command for %s\n", type, modname);
 }
 
-enum exists_response {
-	EXISTS_ERROR,
-	EXISTS_SILENT_ERROR,
-	EXISTS_IGNORE,
-};
-
 /* Actually do the insert.  Frees second arg. */
 static void insmod(struct list_head *list,
 		   char *optstring,
 		   const char *newname,
-		   enum exists_response exists,
+		   int first_time,
 		   errfn_t error,
 		   int dry_run,
 		   int verbose,
@@ -683,7 +718,7 @@ static void insmod(struct list_head *list,
 		   int strip_vermagic,
 		   int strip_modversion)
 {
-	int ret;
+	int ret, fd;
 	unsigned long len;
 	void *map;
 	const char *command;
@@ -692,30 +727,40 @@ static void insmod(struct list_head *list,
 	/* Take us off the list. */
 	list_del(&mod->list);
 
-	/* Do things we depend on first, but don't die if they fail. */
+	/* Do things we (or parent) depend on first, but don't die if
+	 * they fail. */
 	if (!list_empty(list)) {
-		insmod(list, NOFAIL(strdup("")), NULL, EXISTS_IGNORE, warn,
+		insmod(list, NOFAIL(strdup("")), NULL, 0, warn,
 		       dry_run, verbose, options, commands, 0, ignore_proc,
 		       strip_vermagic, strip_modversion);
 	}
 
-	/* Did config file override command or add options? */
+	fd = lock_file(mod->filename);
+	if (fd < 0) {
+		error("Could not open '%s': %s\n",
+		      mod->filename, strerror(errno));
+		goto out_optstring;
+	}
+
+	/* Don't do ANYTHING if already in kernel. */
+	if (!ignore_proc
+	    && module_in_kernel(newname ?: mod->modname, NULL) == 1)
+		goto exists_error;
+
 	command = find_command(mod->modname, commands);
 	if (command && !ignore_commands) {
+		/* It might recurse: unlock. */
+		unlock_file(fd);
 		do_command(mod->modname, command, verbose, dry_run, error,
 			   "install");
 		goto out_optstring;
 	}
 
-	if (!ignore_proc
-	    && module_in_kernel(newname ?: mod->modname, NULL) == 1)
-		goto exists_error;
-
-	map = grab_file(mod->filename, &len);
+	map = grab_fd(fd, &len);
 	if (!map) {
-		error("Could not open '%s': %s\n",
+		error("Could not read '%s': %s\n",
 		      mod->filename, strerror(errno));
-		goto out_optstring;
+		goto out_unlock;
 	}
 
 	/* Rename it? */
@@ -730,8 +775,7 @@ static void insmod(struct list_head *list,
 	/* Config file might have given more options */
 	optstring = add_extra_options(mod->modname, optstring, options);
 
-	if (verbose)
-		printf("insmod %s %s\n", mod->filename, optstring);
+	verbose_printf(verbose, "insmod %s %s\n", mod->filename, optstring);
 
 	if (dry_run)
 		goto out;
@@ -745,123 +789,103 @@ static void insmod(struct list_head *list,
 	}
  out:
 	release_file(map, len);
+ out_unlock:
+	unlock_file(fd);
  out_optstring:
 	free(optstring);
 	return;
 
 exists_error:
-	if (exists == EXISTS_ERROR)
+	if (first_time)
 		error("Module %s already in kernel.\n",
 		      newname ?: mod->modname);
-	else if (exists == EXISTS_SILENT_ERROR)
-		exit(1);
-	goto out_optstring;
+	goto out_unlock;
 }
 
-enum noexists_response {
-	NOEXISTS_ERROR,
-	NOEXISTS_SILENT_ERROR,
-	NOEXISTS_IGNORE,
-};
-
 /* Do recursive removal. */
 static void rmmod(struct list_head *list,
-		  enum noexists_response noexists,
+		  const char *name,
+		  int first_time,
 		  errfn_t error,
 		  int dry_run,
 		  int verbose,
 		  struct module_command *commands,
-		  int ignore_commands)
+		  int ignore_commands,
+		  int ignore_inuse)
 {
 	const char *command;
 	unsigned int usecount = 0;
+	int lock;
 	struct module *mod = list_entry(list->next, struct module, list);
 
 	/* Take first one off the list. */
 	list_del(&mod->list);
 
+	/* Ignore failure; it's best effort here. */
+	lock = lock_file(mod->filename);
+
+	if (!name)
+		name = mod->modname;
+
+	/* Even if renamed, find commands to orig. name. */
 	command = find_command(mod->modname, commands);
 	if (command && !ignore_commands) {
+		/* It might recurse: unlock. */
+		unlock_file(lock);
 		do_command(mod->modname, command, verbose, dry_run, error,
 			   "remove");
-		goto remove_rest;
+		goto remove_rest_no_unlock;
 	}
 
-	if (module_in_kernel(mod->modname, &usecount) == 0)
+	if (module_in_kernel(name, &usecount) == 0)
 		goto nonexistent_module;
 
 	if (usecount != 0) {
-		if (noexists != NOEXISTS_IGNORE)
-			error("Module %s is in use.\n", mod->modname);
+		if (!ignore_inuse)
+			error("Module %s is in use.\n", name);
 		goto remove_rest;
 	}
 
-	if (verbose)
-		printf("rmmod %s\n", mod->filename);
+	verbose_printf(verbose, "rmmod %s\n", mod->filename);
 
 	if (dry_run)
 		goto remove_rest;
 
-	if (delete_module(mod->modname, O_EXCL) != 0) {
+	if (delete_module(name, O_EXCL) != 0) {
 		if (errno == ENOENT)
 			goto nonexistent_module;
 		error("Error removing %s (%s): %s\n",
-		      mod->modname, mod->filename,
+		      name, mod->filename,
 		      remove_moderror(errno));
 	}
 
  remove_rest:
+	unlock_file(lock);
+ remove_rest_no_unlock:
 	/* Now do things we depend. */
 	if (!list_empty(list))
-		rmmod(list, NOEXISTS_IGNORE, warn, dry_run, verbose, commands,
-		      0);
+		rmmod(list, NULL, 0, warn, dry_run, verbose, commands,
+		      0, 1);
 	return;
 
 nonexistent_module:
-	if (noexists == NOEXISTS_ERROR)
+	if (first_time)
 		fatal("Module %s is not in kernel.\n", mod->modname);
-	else if (noexists == NOEXISTS_SILENT_ERROR)
-		exit(1);
 	goto remove_rest;
 }
 
-static int do_modprobe(const char *argv0, const char *modname)
-{
-	int ret;
-
-	ret = fork();
-	if (ret < 0)
-		fatal("Failed to fork to run modprobe %s: %s\n",
-		      modname, strerror(errno));
-	if (ret > 0) {
-		int status;
-		waitpid(ret, &status, 0);
-		if (WIFEXITED(status))
-			ret = (WEXITSTATUS(status) != 0);
-		else ret = 1;
-	} else {
-		execlp(argv0, argv0, modname, NULL);
-		fatal("Failed to run modprobe %s: %s\n",
-		      modname, strerror(errno));
-	}
-
-	return ret;
-}
-
-/* Final dir of path must match subpath. */
+/* Does path contain directory(s) subpath? */
 static int type_matches(const char *path, const char *subpath)
 {
-	char *prev, *slash = strrchr(path, '/');
-	if (!slash)
-		return 0;
+	char *subpath_with_slashes;
+	int ret;
 
-	prev = slash - (strlen(subpath)+1);
-	if (prev < path)
-		return 0;
+	asprintf(&subpath_with_slashes, "/%s/", subpath);
+	NOFAIL(subpath_with_slashes);
 
-	if (prev[0] == '/' && memcmp(prev+1, subpath, strlen(subpath)) == 0)
-		return 1;
-	return 0;
+	ret = (strstr(path, subpath_with_slashes) != NULL);
+	free(subpath_with_slashes);
+	return ret;
 }
 
 static char *underscores(char *string)
@@ -875,27 +899,17 @@ static char *underscores(char *string)
 	return string;
 }
 
-static int do_wildcard(const char *argv0,
-		       const char *dirname,
-		       int list_only,
+static int do_wildcard(const char *dirname,
 		       const char *type,
-		       int all,
-		       int dry_run,
-		       int verbose,
-		       char *wildcard)
+		       const char *wildcard)
 {
 	char modules_dep_name[strlen(dirname) + sizeof("modules.dep") + 1];
-	char *line;
+	char *line, *wcard;
 	FILE *modules_dep;
-	int ret;
-
-	if (list_only || dry_run) /* Always "OK" even if nothing matches */
-		ret = 0;
-	else /* Otherwise, only success of one succeeds. */
-		ret = 1;
 
 	/* Canonicalize wildcard */
-	underscores(wildcard);
+	wcard = strdup(wildcard);
+	underscores(wcard);
 
 	sprintf(modules_dep_name, "%s/%s", dirname, "modules.dep");
 	modules_dep = fopen(modules_dep_name, "r");
@@ -905,7 +919,6 @@ static int do_wildcard(const char *argv0,
 
 	while ((line = getline_wrapped(modules_dep, NULL)) != NULL) {
 		char *ptr;
-		char modname[strlen(line)+1];
 
 		/* Ignore lines without : or which start with a # */
 		ptr = strchr(line, ':');
@@ -913,31 +926,20 @@ static int do_wildcard(const char *argv0,
 			goto next;
 		*ptr = '\0';
 
-		/* "type" must match final directory part. */
-		if (type && !type_matches(line, type))
-			goto next;
+		/* "type" must match complete directory component(s). */
+		if (!type || type_matches(line, type)) {
+			char modname[strlen(line)+1];
 
-		filename2modname(modname, line);
-		if (fnmatch(wildcard, modname, 0) == 0) {
-			if (list_only)
+			filename2modname(modname, line);
+			if (fnmatch(wcard, modname, 0) == 0)
 				printf("%s\n", line);
-			else {
-				if (verbose)
-					printf("modprobe %s\n", modname);
-				if (!dry_run)
-					ret &= do_modprobe(argv0, modname);
-				/* They might not want us to load all. */
-				if (ret == 0 && !all) {
-					free(line);
-					return 0;
-				}
-			}
 		}
 	next:
 		free(line);
 	}
 
-	return ret;
+	free(wcard);
+	return 0;
 }
 
 static char *strsep_skipspace(char **string, char *delim)
@@ -948,30 +950,31 @@ static char *strsep_skipspace(char **string, char *delim)
 	return strsep(string, delim);
 }
 
-/* FIXME: Maybe should be extended to "alias a b [and|or c]...". --RR */
+/* Recursion */
+static int read_config(const char *filename,
+		       const char *name,
+		       int dump_only,
+		       int removing,
+		       struct module_options **options,
+		       struct module_command **commands,
+		       char **alias);
 
-/* Simple format, ignore lines starting with #, one command per line.
-   Returns NULL or resolved alias. */
-static char *read_config(const char *filename,
-			 int mustload,
-			 const char *name,
-			 int dump_only,
-			 int removing,
-			 struct module_options **options,
-			 struct module_command **commands)
+/* FIXME: Maybe should be extended to "alias a b [and|or c]...". --RR */
+static int read_config_file(const char *filename,
+			    const char *name,
+			    int dump_only,
+			    int removing,
+			    struct module_options **options,
+			    struct module_command **commands,
+			    char **alias)
 {
-	FILE *cfile;
 	char *line;
-	char *result = NULL;
 	unsigned int linenum = 0;
+	FILE *cfile;
 
 	cfile = fopen(filename, "r");
-	if (!cfile) {
-		if (mustload)
-			fatal("Failed to open config file %s: %s\n",
-			      filename, strerror(errno));
-		return NULL;
-	}
+	if (!cfile)
+		return 0;
 
 	while ((line = getline_wrapped(cfile, &linenum)) != NULL) {
 		char *ptr = line;
@@ -993,21 +996,25 @@ static char *read_config(const char *filename,
 			if (!wildcard || !realname)
 				grammar(cmd, filename, linenum);
 			else if (fnmatch(wildcard,name,0) == 0)
-				result = NOFAIL(strdup(realname));
+				*alias = NOFAIL(strdup(realname));
 		} else if (strcmp(cmd, "include") == 0) {
-			char *newresult, *newfilename;
+			char *newalias = NULL, *newfilename;
 
 			newfilename = strsep_skipspace(&ptr, "\t ");
 			if (!newfilename)
 				grammar(cmd, filename, linenum);
 			else {
-				newresult = read_config(newfilename, 1, name,
-							dump_only, removing,
-							options, commands);
+				if (!read_config(newfilename, name,
+						 dump_only, removing,
+						 options, commands, &newalias))
+					warn("Failed to open included"
+					      " config file %s: %s\n",
+					      newfilename, strerror(errno));
+
 				/* Files included override aliases,
 				   etc that was already set ... */
-				if (newresult)
-					result = newresult;
+				if (newalias)
+					*alias = newalias;
 			}
 		} else if (strcmp(cmd, "options") == 0) {
 			modname = strsep_skipspace(&ptr, "\t ");
@@ -1042,8 +1049,78 @@ static char *read_config(const char *filename,
 		free(line);
 	}
 	fclose(cfile);
+	return 1;
+}
 
-	return result;
+/* Simple format, ignore lines starting with #, one command per line.
+   Returns NULL or resolved alias. */
+static int read_config(const char *filename,
+		       const char *name,
+		       int dump_only,
+		       int removing,
+		       struct module_options **options,
+		       struct module_command **commands,
+		       char **alias)
+{
+	DIR *dir;
+
+	/* If it's a directory, recurse. */
+	dir = opendir(filename);
+	if (dir) {
+		struct dirent *i;
+
+		while ((i = readdir(dir)) != NULL) {
+			if (!streq(i->d_name,".") && !streq(i->d_name,"..")) {
+				char sub[strlen(filename) + 1
+					 + strlen(i->d_name) + 1];
+
+				sprintf(sub, "%s/%s", filename, i->d_name);
+				if (!read_config(sub, name,
+						 dump_only, removing,
+						 options, commands, alias))
+					warn("Failed to open"
+					      " config file %s: %s\n",
+					      sub, strerror(errno));
+			}
+		}
+		closedir(dir);
+		return 1;
+	}
+
+	return read_config_file(filename, name, dump_only, removing,
+				options, commands, alias);
+}
+
+static const char *default_configs[] = 
+{
+	"/etc/modprobe.conf",
+	"/etc/modprobe.d",
+};
+
+static void read_toplevel_config(const char *filename,
+				 const char *name,
+				 int dump_only,
+				 int removing,
+				 struct module_options **options,
+				 struct module_command **commands,
+				 char **alias)
+{
+	unsigned int i;
+
+	if (filename) {
+		if (!read_config(filename, name, dump_only, removing,
+				 options, commands, alias))
+			fatal("Failed to open config file %s: %s\n",
+			      filename, strerror(errno));
+		return;
+	}
+
+	/* Try defaults. */
+	for (i = 0; i < ARRAY_SIZE(default_configs); i++) {
+		if (read_config(default_configs[i], name, dump_only,
+				removing, options, commands, alias))
+			return;
+	}
 }
 
 static void add_to_env_var(const char *option)
@@ -1073,7 +1150,7 @@ static char **merge_args(char *args, char *argv[], int *argc)
 		num_env++;
 		newargs = NOFAIL(realloc(newargs,
 					 sizeof(newargs[0])
-					 * (num_env + *argc)));
+					 * (num_env + *argc + 1)));
 		newargs[num_env] = arg;
 	}
 
@@ -1092,12 +1169,14 @@ static char *gather_options(char *argv[])
 
 	/* Rest is module options */
 	while (*argv) {
-		if (strchr(*argv, ' ')) {
-			/* Spaces handled by "" pairs, but no way of
-			   escaping quotes */
-			char protected_option[strlen(optstring) + 3];
-			sprintf(protected_option, "\"%s\"", *argv);
-			optstring = append_option(optstring, protected_option);
+		/* Quote value if it contains spaces. */
+		unsigned int eq = strcspn(*argv, "=");
+
+		if (strchr(*argv+eq, ' ') && !strchr(*argv, '"')) {
+			char quoted[strlen(*argv) + 3];
+			(*argv)[eq] = '\0';
+			sprintf(quoted, "%s=\"%s\"", *argv, *argv+eq+1);
+			optstring = append_option(optstring, quoted);
 		} else
 			optstring = append_option(optstring, *argv);
 		argv++;
@@ -1114,7 +1193,6 @@ static struct option options[] = { { "verbose", 0, NULL, 'v' },
 				   { "showconfig", 0, NULL, 'c' },
 				   { "autoclean", 0, NULL, 'k' },
 				   { "quiet", 0, NULL, 'q' },
-				   { "really-quiet", 0, NULL, 'Q' },
 				   { "show", 0, NULL, 'n' },
 				   { "dry-run", 0, NULL, 'n' },
 				   { "syslog", 0, NULL, 's' },
@@ -1124,15 +1202,30 @@ static struct option options[] = { { "verbose", 0, NULL, 'v' },
 				   { "ignore-install", 0, NULL, 'i' },
 				   { "ignore-remove", 0, NULL, 'i' },
 				   { "force", 0, NULL, 'f' },
-				   { "force-vermagic", 0, NULL, '0' },
-				   { "force-modversion", 0, NULL, '1' },
+				   { "force-vermagic", 0, NULL, 1 },
+				   { "force-modversion", 0, NULL, 2 },
 				   { "set-version", 1, NULL, 'S' },
 				   { "show-depends", 0, NULL, 'D' },
+				   { "first-time", 0, NULL, 3 },
 				   { NULL, 0, NULL, 0 } };
 
-#define DEFAULT_CONFIG "/etc/modprobe.conf"
 #define MODPROBE_DEVFSD_CONF "/etc/modprobe.devfs"
 
+/* This is a horrible hack to allow devfsd, which calls modprobe with
+   -C /etc/modules.conf or /etc/modules.devfs, to work.  FIXME. */
+/* Modern devfsd or variants should use -q explicitly in 2.6. */
+static int is_devfs_call(char *argv[])
+{
+	unsigned int i;
+
+	/* Look for "/dev" arg */
+	for (i = 1; argv[i]; i++) {
+		if (strncmp(argv[i], "/dev/", 5) == 0)
+			return 1;
+	}
+	return 0;
+}
+
 int main(int argc, char *argv[])
 {
 	struct utsname buf;
@@ -1142,20 +1235,21 @@ int main(int argc, char *argv[])
 	int dry_run = 0;
 	int remove = 0;
 	int verbose = 0;
+	int unknown_silent = 0;
 	int list_only = 0;
 	int all = 0;
 	int ignore_commands = 0;
 	int strip_vermagic = 0;
 	int strip_modversion = 0;
 	int ignore_proc = 0;
-	enum noexists_response noexists = NOEXISTS_ERROR;
-	enum exists_response exists = EXISTS_ERROR;
+	int first_time = 0;
 	unsigned int i, num_modules;
 	char *type = NULL;
 	const char *config = NULL, *command = NULL;
 	char *dirname, *optstring;
-	char *modname, *newname = NULL;
+	char *newname = NULL;
 	char *aliasfilename, *symfilename;
+	errfn_t error = fatal;
 
 	/* Prepend options from environment. */
 	argv = merge_args(getenv("MODPROBE_OPTIONS"), argv, &argc);
@@ -1184,9 +1278,31 @@ int main(int argc, char *argv[])
 			buf.release[sizeof(buf.release)-1] = '\0';
 			break;
 		case 'C':
-			add_to_env_var("-C");
-			add_to_env_var(optarg);
-			config = optarg;
+			if (is_devfs_call(argv)) {
+				if (streq("/etc/modules.devfs", optarg)) {
+					config = MODPROBE_DEVFSD_CONF;
+					add_to_env_var("-C");
+					add_to_env_var(config);
+					/* Fall thru to -q */
+				} else if (streq("/etc/modules.conf", optarg))
+					/* Ignore config, fall thru to -q */
+					;
+				else {
+					/* False alarm.  Treat as normal. */
+					config = optarg;
+					add_to_env_var("-C");
+					add_to_env_var(config);
+					break;
+				}
+			} else {
+				config = optarg;
+				add_to_env_var("-C");
+				add_to_env_var(config);
+				break;
+			}
+		case 'q':
+			unknown_silent = 1;
+			add_to_env_var("-q");
 			break;
 		case 'D':
 			dry_run = 1;
@@ -1211,6 +1327,7 @@ int main(int argc, char *argv[])
 			break;
 		case 'a':
 			all = 1;
+			error = warn;
 			break;
 		case 'k':
 			/* FIXME: This should actually do something */
@@ -1218,20 +1335,6 @@ int main(int argc, char *argv[])
 		case 'n':
 			dry_run = 1;
 			break;
-		case 'q':
-			/* install commands can rely on failure status
-			 * of children, so we only suppress output.
-			 *  install ip_conntrack /sbin/modprobe ip_conntrack
-			 *  && { /sbin/modprobe ip_conntrack_ftp; /bin/true; }
-			 */
-			add_to_env_var("-q");
-			noexists = NOEXISTS_SILENT_ERROR;
-			exists = EXISTS_SILENT_ERROR;
-			break;
-		case 'Q':
-			add_to_env_var("-Q");
-			quiet = 1;
-			break;
 		case 's':
 			add_to_env_var("-s");
 			log = 1;
@@ -1243,12 +1346,15 @@ int main(int argc, char *argv[])
 			strip_vermagic = 1;
 			strip_modversion = 1;
 			break;
-		case '0':
+		case 1:
 			strip_vermagic = 1;
 			break;
-		case '1':
+		case 2:
 			strip_modversion = 1;
 			break;
+		case 3:
+			first_time = 1;
+			break;
 		default:
 			print_usage(argv[0]);
 		}
@@ -1272,70 +1378,56 @@ int main(int argc, char *argv[])
 				    + sizeof("/modules.symbols")));
 	sprintf(symfilename, "%s/modules.symbols", dirname);
 
-	/* Old-style -t xxx wildcard?  Or -l, or -a. */
-	if (all || type || list_only) {
+	/* Old-style -t xxx wildcard?  Only with -l. */
+	if (list_only) {
+		if (optind+1 < argc)
+			fatal("Can't have multiple wildcards\n");
 		/* fprintf(stderr, "man find\n"); return 1; */
-		return do_wildcard(argv[0], dirname, list_only, type, all,
-				   dry_run, verbose, argv[optind] ?: "*");
+		return do_wildcard(dirname, type, argv[optind]?:"*");
+	}
+	if (type)
+		fatal("-t only supported with -l");
+
+	if (dump_only) {
+		struct module_command *commands = NULL;
+		struct module_options *modoptions = NULL;
+		char *a = NULL;
+
+		read_toplevel_config(config, "", 1, 0,
+				     &modoptions, &commands, &a);
+		read_config(aliasfilename, "", 1, 0,&modoptions, &commands,&a);
+		read_config(symfilename, "", 1, 0, &modoptions, &commands, &a);
+		exit(0);
 	}
 
-	if (remove) {
+	if (remove || all) {
 		num_modules = argc - optind;
 		optstring = NOFAIL(strdup(""));
-		/* -r only allows certain restricted options */
-		if (newname)
-			fatal("Can't specify replacement name with -r\n");
 	} else {
 		num_modules = 1;
 		optstring = gather_options(argv+optind+1);
 	}
 
- 	/* This is a horrible hack to allow devfsd, which calls
-	   modprobe with -C /etc/modules.conf or /etc/modules.devfs,
-	   to work.  FIXME. */
-	/* Modern devfsd or variants should use -Q explicitly in 2.6. */
-	if (config && !dump_only && strncmp(argv[optind], "/dev/", 5) == 0) {
-		if (strcmp("/etc/modules.conf", config) == 0) {
-			quiet = 1;
-			config = NULL;
-		} else if (strcmp("/etc/modules.devfs", config) == 0) {
-			config = MODPROBE_DEVFSD_CONF;
-			quiet = 1;
-		}
-	}
-
-	if (dump_only) {
-		struct module_command *commands = NULL;
-		struct module_options *modoptions = NULL;
-
-		read_config(config ?: DEFAULT_CONFIG,
-			    config ? 1 : 0, "", 1, 0, &modoptions, &commands);
-		read_config(aliasfilename, 0, "", 1, 0,&modoptions, &commands);
-		read_config(symfilename, 0, "", 1, 0, &modoptions, &commands);
-		exit(0);
-	}
-
-	/* num_modules is always 1 except for remove */
+	/* num_modules is always 1 except for -r or -a. */
 	for (i = 0; i < num_modules; i++) {
 		struct module_command *commands = NULL;
 		struct module_options *modoptions = NULL;
 		LIST_HEAD(list);
 		char *modulearg = argv[optind + i];
+		char *modname = NULL;
 
 		/* Convert name we are looking for */
 		underscores(modulearg);
 
 		/* Returns the resolved alias, options */
-		modname = read_config(config ?: DEFAULT_CONFIG,
-				      config ? 1 : 0,
-				      modulearg, 0,
-				      remove, &modoptions, &commands);
+		read_toplevel_config(config, modulearg, 0,
+				     remove, &modoptions, &commands, &modname);
 
 		/* No luck?  Try symbol names, if starts with symbol:. */
 		if (!modname
 		    && strncmp(modulearg, "symbol:", strlen("symbol:") == 0))
-			modname = read_config(symfilename, 0, modulearg, 0,
-					      remove, &modoptions, &commands);
+			read_config(symfilename, modulearg, 0,
+				    remove, &modoptions, &commands, &modname);
 
 		/* If we have an alias, gather any options associated with it
 		   (needs to happen after parsing complete). */
@@ -1346,13 +1438,16 @@ int main(int argc, char *argv[])
 			read_depends(dirname, modname, &list);
 		} else {
 			read_depends(dirname, modulearg, &list);
-			/* We don't allow canned aliases to override real
-			   modules, so we delay lookup until now. */
+			/* We don't allow canned aliases to override
+			   real modules, or install commands, so we
+			   delay lookup until now. */
 			if (list_empty(&list)
-			    && (modname = read_config(aliasfilename, 0,
-						      modulearg, 0,
-						      remove, &modoptions,
-						      &commands)))
+			    && !find_command(modulearg, commands)
+			    && read_config(aliasfilename,
+					   modulearg, 0,
+					   remove, &modoptions,
+					   &commands, &modname)
+			    && modname)
 				goto got_modname;
 
 			modname = strdup(modulearg);
@@ -1367,19 +1462,19 @@ int main(int argc, char *argv[])
 					   fatal, remove ? "remove":"install");
 				continue;
 			}
-			if (noexists == NOEXISTS_ERROR)
-				fatal("Module %s not found.\n", modname);
-			else
+			if (unknown_silent)
 				exit(1);
+			error("Module %s not found.\n", modname);
+			continue;
 		}
 
 		if (remove)
-			rmmod(&list, noexists, fatal, dry_run, verbose,
-			      commands, ignore_commands);
+			rmmod(&list, newname, first_time, error, dry_run,
+			      verbose, commands, ignore_commands, 0);
 		else
-			insmod(&list, optstring, newname, exists, fatal,
-			       dry_run, verbose, modoptions, commands,
-			       ignore_commands, ignore_proc,
+			insmod(&list, NOFAIL(strdup(optstring)), newname,
+			       first_time, error, dry_run, verbose, modoptions,
+			       commands, ignore_commands, ignore_proc,
 			       strip_vermagic, strip_modversion);
 		free(modname);
 	}
@@ -1387,6 +1482,7 @@ int main(int argc, char *argv[])
 	free(dirname);
 	free(aliasfilename);
 	free(symfilename);
+	free(optstring);
 	if (log)
 		closelog();
 
