@@ -29,7 +29,7 @@ static void PERBIT(load_symbols)(struct module *module)
 	unsigned long i, size;
 
 	/* New-style: strings are in this section. */
-	ksymstrings = PERBIT(load_section)(module->mmap, "__ksymtab_strings",
+	ksymstrings = PERBIT(load_section)(module->data, "__ksymtab_strings",
 					   &size);
 	if (ksymstrings) {
 		unsigned int i = 0;
@@ -42,7 +42,7 @@ static void PERBIT(load_symbols)(struct module *module)
 			i += strlen(ksymstrings+i);
 		}
 		/* GPL symbols too */
-		ksymstrings = PERBIT(load_section)(module->mmap,
+		ksymstrings = PERBIT(load_section)(module->data,
 						   "__ksymtab_strings_gpl",
 						   &size);
 		for (;;) {
@@ -57,24 +57,28 @@ static void PERBIT(load_symbols)(struct module *module)
 	}
 
 	/* Old-style. */
-	ksyms = PERBIT(load_section)(module->mmap, "__ksymtab", &size);
+	ksyms = PERBIT(load_section)(module->data, "__ksymtab", &size);
 	for (i = 0; i < size / sizeof(struct PERBIT(kernel_symbol)); i++)
 		add_symbol(ksyms[i].name, module);
-	ksyms = PERBIT(load_section)(module->mmap, "__gpl_ksymtab", &size);
+	ksyms = PERBIT(load_section)(module->data, "__gpl_ksymtab", &size);
 	for (i = 0; i < size / sizeof(struct PERBIT(kernel_symbol)); i++)
 		add_symbol(ksyms[i].name, module);
 }
 
 static char *PERBIT(get_aliases)(struct module *module, unsigned long *size)
 {
-	return PERBIT(load_section)(module->mmap, ".modalias", size);
+	return PERBIT(load_section)(module->data, ".modalias", size);
 }
 
 static char *PERBIT(get_modinfo)(struct module *module, unsigned long *size)
 {
-	return PERBIT(load_section)(module->mmap, ".modinfo", size);
+	return PERBIT(load_section)(module->data, ".modinfo", size);
 }
 
+#ifndef STT_REGISTER
+#define STT_REGISTER    13              /* Global register reserved to app. */
+#endif
+
 /* Calculate the dependencies for this module */
 static void PERBIT(calculate_deps)(struct module *module, int verbose)
 {
@@ -85,8 +89,8 @@ static void PERBIT(calculate_deps)(struct module *module, int verbose)
 	ElfPERBIT(Ehdr) *hdr;
 	int handle_register_symbols;
 
-	strings = PERBIT(load_section)(module->mmap, ".strtab", &size);
-	syms = PERBIT(load_section)(module->mmap, ".symtab", &size);
+	strings = PERBIT(load_section)(module->data, ".strtab", &size);
+	syms = PERBIT(load_section)(module->data, ".symtab", &size);
 
 	if (!strings || !syms) {
 		warn("Couldn't find symtab and strtab in module %s\n",
@@ -94,7 +98,7 @@ static void PERBIT(calculate_deps)(struct module *module, int verbose)
 		return;
 	}
 
-	hdr = module->mmap;
+	hdr = module->data;
 	handle_register_symbols = 0;
 	if (hdr->e_machine == EM_SPARC ||
 	    hdr->e_machine == EM_SPARCV9)
@@ -111,7 +115,7 @@ static void PERBIT(calculate_deps)(struct module *module, int verbose)
 
 			/* Not really undefined: sparc gcc 3.3 creates
                            U references when you have global asm
-                           variables, to avoid anyone else mising
+                           variables, to avoid anyone else misusing
                            them. */
 			if (handle_register_symbols
 			    && (ELFPERBIT(ST_TYPE)(syms[i].st_info)
@@ -149,6 +153,10 @@ static void *PERBIT(deref_sym)(ElfPERBIT(Ehdr) *hdr, const char *name)
 
 	for (i = 0; i < size / sizeof(syms[0]); i++) {
 		if (strcmp(strings + syms[i].st_name, name) == 0) {
+			/* In BSS?  Happens for empty device tables on
+			 * recent GCC versions. */
+			if (sechdrs[syms[i].st_shndx].sh_type == SHT_NOBITS)
+				return NULL;
 			return (void *)hdr
 				+ sechdrs[syms[i].st_shndx].sh_offset
 				+ syms[i].st_value;
@@ -161,20 +169,33 @@ static void *PERBIT(deref_sym)(ElfPERBIT(Ehdr) *hdr, const char *name)
 static void PERBIT(fetch_tables)(struct module *module)
 {
 	module->pci_size = PERBIT(PCI_DEVICE_SIZE);
-	module->pci_table = PERBIT(deref_sym)(module->mmap,
+	module->pci_table = PERBIT(deref_sym)(module->data,
 					"__mod_pci_device_table");
 
 	module->usb_size = PERBIT(USB_DEVICE_SIZE);
-	module->usb_table = PERBIT(deref_sym)(module->mmap,
+	module->usb_table = PERBIT(deref_sym)(module->data,
 					"__mod_usb_device_table");
 
 	module->ccw_size = PERBIT(CCW_DEVICE_SIZE);
-	module->ccw_table = PERBIT(deref_sym)(module->mmap,
+	module->ccw_table = PERBIT(deref_sym)(module->data,
 					"__mod_ccw_device_table");
 
 	module->ieee1394_size = PERBIT(IEEE1394_DEVICE_SIZE);
-	module->ieee1394_table = PERBIT(deref_sym)(module->mmap,
+	module->ieee1394_table = PERBIT(deref_sym)(module->data,
 					"__mod_ieee1394_device_table");
+
+	module->pnp_size = PERBIT(PNP_DEVICE_SIZE);
+	module->pnp_table = PERBIT(deref_sym)(module->data,
+					"__mod_pnp_device_table");
+
+	module->pnp_card_size = PERBIT(PNP_CARD_DEVICE_SIZE);
+	module->pnp_card_table = PERBIT(deref_sym)(module->data,
+					"__mod_pnp_card_device_table");
+	module->pnp_card_offset = PERBIT(PNP_CARD_DEVICE_OFFSET);
+
+	module->input_size = PERBIT(INPUT_DEVICE_SIZE);
+	module->input_table = PERBIT(deref_sym)(module->data,
+					"__mod_input_device_table");
 }
 
 struct module_ops PERBIT(mod_ops) = {
